/* Theme backgrounds */
window.postmodern {
    background: #e8e9f2;
}

window.smooth {
    background: #eeede3;
}

window.educational {
    background: #faf4f4;
}

/* Scores */
columnview {
    background: transparent;
}

columnview header button,
columnview row cell {
    padding-left: 12px;
    padding-right: 12px;
}
